#!/usr/bin/perl

use strict;
use warnings;
use Data::Dumper;
use Statistics::Descriptive;

# Set to 1 for CSV stats, 0 for table of means used in the report.
my $stats = 1;
my %accesses;

my $interval = undef;
while(<>)
{
    chomp;
    my $access = parse_line( $_ );
    next if $access->{agent} =~ /Java/;
    next unless $access->{request} =~ m{/logit\.html};

    my $query = (split( /\?/, $access->{request} ))[1];
    my %parms = map { s/%([\da-f][\da-f])/chr hex $1/eg; $_; }
                ($query =~ /([^&=]+)=([^&]+)/g);
    my @fields = qw/count avg elapsed tot_count tot_avg tot_elapsed/;
    @{$access}{@fields} = @parms{@fields};
    @{$access}{qw/app interval/} = ($1,$2) if $parms{marker} =~ /(\w+)\s+(\d+)/;
    delete $access->{request};
    push @{$accesses{$access->{app}}->{$access->{agent}}->{$access->{interval}}}, $access;
}

if( $stats )
{
    foreach my $app (sort keys %accesses)
    {
        foreach my $agent (sort keys %{$accesses{$app}})
        {
            my $summary = $accesses{$app}->{$agent};
            print "$app: $agent:\n";
            my @stat_fields = qw/mean median stddev min max/;
            print join( "\t", 'interval', (map { "sample_$_" } @stat_fields),
                                        (map { "total_$_" } @stat_fields)
            ), "\n";
            foreach my $interval (sort { $a <=> $b } keys %{$summary})
            {
                print "$interval\t";
                my $stat = calc_stats( 'avg', $summary->{$interval} );
                print join( "\t", map { sprintf '%.4f', $_ } @{$stat}{@stat_fields} );
                $stat = calc_stats( 'tot_avg', $summary->{$interval} );
                print "\t", join( "\t", map { sprintf '%.4f', $_ } @{$stat}{@stat_fields} ), "\n";
            }
            print "\n";
        }
    }
}
else
{
    my @agents;
    foreach my $app (sort keys %accesses)
    {
        print "$app:\n";
        my %table = ();
        @agents = sort keys %{$accesses{$app}} unless @agents;
        foreach my $agent (@agents)
        {
            my $summary = $accesses{$app}->{$agent};
            foreach my $interval (sort { $a <=> $b } keys %{$summary})
            {
                my $stat = calc_stats( 'avg', $summary->{$interval} );
                $table{$interval}->{$agent} = sprintf '%.02f', $stat->{mean};
            }
        }
        print "\t", join( "\t", @agents ), "\n";
        foreach my $i (sort { $a <=> $b } keys %table) {
            print "$i\t", join( "\t", @{$table{$i}}{@agents} ), "\n";
        }
    }
}

sub calc_stats
{
    my $which = shift;
    my @samples = @{ shift() };
    my %stats;
    my @avgs = map { $_->{$which} } @samples;

    my $stat = Statistics::Descriptive::Full->new();
#    $stat->add_data( @avgs[0..49] );
#    $stat->add_data( @avgs[1..50] );
    $stat->add_data( (sort { $a <=> $b } @avgs)[1..50] );
    $stats{'median'} = $stat->median();
    $stats{'stddev'} = $stat->standard_deviation();
    $stats{'mean'} = $stat->mean();
    $stats{'min'} = $stat->min();
    $stats{'max'} = $stat->max();

    return \%stats;
}


sub parse_line
{
    my $line = shift;

    my @fields = grep { defined $_ } ($line =~ m{ \[([^\]]+)\] | "([^"]+)" | (\S+) }gx);

    my $access = {
        time => $fields[3],
        request => $fields[4],
        agent => $fields[8]
    };

    $access->{agent} = $1 if $access->{agent} =~ m{((?:Firefox|Opera|Batik)/[\d\.]+|Batik/ batik-[\d\._]+)};
    $access->{time} = $1 if $access->{time} =~ m{:(\d\d:\d\d:\d\d)};
    $access->{request} =~ s/^(GET|POST) //;
    $access->{request} =~ s{ HTTP/[\d\.]+$}{};

    return $access;
}


