#!/usr/bin/perl

use strict;
use warnings;
use Carp;
use Statistics::Descriptive;
use Getopt::Long;
use Pod::Usage;
use List::Util;
use Template;

my $VERSION = '0.0.1';

my %options = (
    'outfile' => undef,
    'app' => undef,
);

my @yranges = (
    [  10,  1 ],
    [  25,  5 ],
    [  50,  5 ],
    [ 100, 10 ],
    [ 150, 15 ],
    [ 200, 20 ],
    [ 250, 25 ],
    [ 300, 30 ],
);

GetOptions(
    'outfile|out=s' => \$options{outfile},
    'app=s' => \$options{app},
) or pod2usage(2);

Carp::croak 'Missing SVG application name' unless $options{app};

my %accesses;

{
    package Dims;
    sub new
    {
        my $class = shift;
        my $self = {
            width => 100, height => 100,
            lmargin => 10, rmargin => 10,
            tmargin => 10, bmargin => 10,
            high => 50,
            @_
        };

        return bless $self, $class;
    }
    sub width   { my $self = shift; return $self->{width}; }
    sub height  { my $self = shift; return $self->{height}; }
    sub top     { my $self = shift; return $self->{tmargin}; }
    sub right   { my $self = shift; return $self->{width}-$self->{rmargin}; }
    sub bottom  { my $self = shift; return $self->{height}-$self->{bmargin}; }
    sub left    { my $self = shift; return $self->{lmargin}; }
    sub high    { my $self = shift; return $self->{high}; }
    sub gwidth  { my $self = shift; return $self->width-$self->{lmargin}-$self->{rmargin}; }
    sub gheight { my $self = shift; return $self->height-$self->{tmargin}-$self->{bmargin}; }
    sub range
    {
        my $self = shift;
        unless( exists $self->{range} ) {
            my $range = List::Util::first { $self->high < $_->[0] } @yranges;
            $self->{range} = $range->[0];
        }
        return $self->{range}
    }
    sub xpoints
    {
        my $self = shift;
        my @points = @_;
        return () unless @points > 1;

        my $x = $self->left;
        my $inc = $self->gwidth/(@points-1);
        my @xpoints;
        foreach my $p ( @points )
        {
            push @xpoints, { p=>$p, x=> clean_num( $x ) };
            $x += $inc;
        }
        return @xpoints;
    }
    sub x_coords
    {
        my $self = shift;
        my $len = shift;
        return () unless $len > 1;

        my $x = $self->left;
        my $inc = $self->gwidth/($len-1);
        my @xpoints;
        foreach my $p ( 1 .. $len )
        {
            push @xpoints, clean_num( $x );
            $x += $inc;
        }
        return @xpoints;
    }
    sub ypoints
    {
        my $self = shift;
        my @points = ();
        my $range = List::Util::first { $self->high < $_->[0] } @yranges;
        for( my $i = 0; $i <= $range->[0]; $i += $range->[1] ) {
            push @points, $i;
        }

        my $y = $self->bottom;
        my $inc = $self->gheight/(@points-1);
        my @ypoints;
        foreach my $l ( @points )
        {
            push @ypoints, { l=>$l, y=>$y };
            $y -= $inc;
        }
        return @ypoints;
    }
    sub x_coord
    {
        my ($self, $i, $num) = @_;

        my $x = $self->left;
        $x += $self->gwidth/($num-1) if $num > 0;

        return $;
    }
    sub clean_num
    {
        my $num = sprintf '%.02f', shift();
        $num =~ s/\.?0+$//;
        return $num;
    }
    sub legend_size
    {
        my ($self) = @_;
        return {
            x => $self->width - 150,
            y => 50,
            width => 105,
            height => (4+$self->{num_lines}*15),
        };
    }
    sub legend
    {
        my ($self, $i) = @_;
        return {
            block=>5 + $i*15,
            text => 14 + $i*15,
        };
    }
}

my $interval = undef;
while(<>)
{
    chomp;
    my $access = parse_line( $_ );
    next if $access->{agent} =~ /Java/;
    next unless $access->{request} =~ m{/logit\.html};

    my $query = (split( /\?/, $access->{request} ))[1];
    my %parms = map { s/%([\da-f][\da-f])/chr hex $1/eg; $_; }
                ($query =~ /([^&=]+)=([^&]+)/g);
    my @fields = qw/count avg elapsed tot_count tot_avg tot_elapsed/;
    @{$access}{@fields} = @parms{@fields};
    @{$access}{qw/app interval/} = ($1,$2) if $parms{marker} =~ /(\w+)\s+(\d+)/;
    next unless $access->{app} eq $options{app};

    delete $access->{request};
    push @{$accesses{$access->{agent}}->{$access->{interval}}}, $access;
}

Carp::croak "No entries for $options{app} were found." unless keys %accesses;

my $tt = Template->new( {POST_CHOMP => 1, EVAL_PERL => 0} )
    or Carp::croak Template->error();

my %results = ();
foreach my $agent (sort keys %accesses)
{
    my $summary = $accesses{$agent};
    my @stat_fields = qw/mean median stddev min max/;

    my $agent_d = $agent;
    $agent_d =~ tr{/}{ };
    foreach my $interval (sort { $a <=> $b } keys %{$summary})
    {
        my $stat = calc_stats( 'avg', $summary->{$interval} );
        $results{$agent_d}->{$interval} = $stat;
    }
}

my @intervals = sort {$a<=>$b} keys %{(values %results)[0]};
my $high = List::Util::max( map { $_->{max} } map { values %{$_} } values %results );

my $dims = Dims->new( 
   width=>600, height=>450,
   lmargin => 40, rmargin =>25,
   tmargin => 25, bmargin => 35,
   high => $high, num_lines => scalar( keys %results ),
);

# Rescale the data to the graph.
foreach my $agent ( keys %results )
{
    foreach my $interval ( keys %{$results{$agent}} )
    {
        my $stats = $results{$agent}->{$interval};
        $stats->{gmean} = scale_it( $stats->{mean} );
        $stats->{gmin} = scale_it( $stats->{min} );
        $stats->{gmax} = scale_it( $stats->{max} );
        $stats->{gsdmax} = scale_it( $stats->{mean}+$stats->{stddev} );
        $stats->{gsd2} = abs( scale_it( 2*$stats->{stddev} ) );
    }
}

sub scale_it
{
    my $num = shift;
    return sprintf( '%.02f', -$num/$dims->range * $dims->gheight );
}

sub data_line {
    my $agent = shift;
    my @vals = $dims->xpoints( @{$results{$agent}}{@intervals} );

    return 'M'.join( ' L', map { "$_->{x},$_->{p}->{gmean}" } @vals );
}

sub data_detail {
    my ($agent) = @_;
    my @details;
    my $result = $results{$agent};
    my $len = @intervals;

    my $top_row = sprintf '%.02f', -$dims->gheight * 0.75;
    my $i = 0;
    foreach my $x ( $dims->x_coords( $len ) )
    {
        my $yd = ($i&1) ? $top_row+70 : $top_row;
        my $xd = $i
            ? (($i+1 == $len) ? $x-70 : $x-35)
            : $x+10;
        my $detail = { x=>$xd, y=>$yd, width=>70, height=>63 };

        my $stat = $result->{$intervals[$i]};
        $detail->{interval} = $intervals[$i];
        $detail->{mean} = sprintf '%.02f', $stat->{mean};
        $detail->{min} = sprintf '%.02f', $stat->{min};
        $detail->{max} = sprintf '%.02f', $stat->{max};
        $detail->{stddev} = sprintf '%.02f', $stat->{stddev};
        push @details, $detail;
        ++$i;
    }

    return @details;
}

sub error_bars {
    my ($agent) = @_;
    my $len = @intervals;

    my @bars = ();
    my $i = 0;
    foreach my $x ( $dims->x_coords( $len ) )
    {
        my $stat = $results{$agent}->{$intervals[$i]};
        my $diff = sprintf( '%.02f', $stat->{gmin}-$stat->{gmax} );
        my $sd = sprintf( '%.02f', $stat->{gsd2}/2 );
        my $bar = "M$x,$stat->{gmax}m-0.2,0 v${diff}m0.4,0v-${diff}";
        $bar .= "M$x,$stat->{gsdmax}h-3v$stat->{gsd2}h6v-$stat->{gsd2}h-3";
        $bar .= "h-3v${sd}h6v-${sd}h-3";
        push @bars, $bar;
        ++$i;
    }

    return @bars;
}

my $vars = {
    height => $dims->height,
    width => $dims->width,
    filename => ($options{outfile} || '<stdout>'),
    app => $options{app},
    heading => { x=>$dims->width/2, y=>15 },
    orig => { x=>$dims->left, y=>$dims->bottom },
    top => { x=>$dims->right, y=>$dims->top },
    xtitle => { x=>$dims->width/2, y=>$dims->height-10 },
    ytitle => { x=>12, y=>$dims->height/2 },
    xpoints => [ $dims->xpoints( @intervals ) ],
    xlabel_y => $dims->bottom+13,
    ypoints => [ $dims->ypoints( $high ) ],
    ylabel_x => $dims->left-5,
    data_pts => sub { $dims->xpoints( @{$results{$_[0]}}{@intervals} ) },
    data_line => \&data_line,
    error_bars => \&error_bars,
    agent => sub { return (sort keys %results)[$_[0]] },
    agent_indexes => [ 0 .. (keys %results)-1 ],
    agent_count => sub { return scalar keys %results; },
    color => sub { return ('blue', 'green', 'red', 'magenta', '#088', 'orange', '#000')[$_[0]]; },
    legend_size => sub { return $dims->legend_size( $_[0] ); },
    legend => sub { return $dims->legend( $_[0] ); },
    data_detail => \&data_detail,
};

$tt->process( \*DATA, $vars, ($options{outfile} ? $options{outfile} : ()) )
    or Carp::croak $tt->error();


sub calc_stats
{
    my $which = shift;
    my @samples = @{ shift() };
    my %stats;
    my @avgs = map { $_->{$which} } @samples;

    my $stat = Statistics::Descriptive::Full->new();
#    $stat->add_data( @avgs[0..49] );
#    $stat->add_data( @avgs[1..50] );
    $stat->add_data( (sort { $a <=> $b } @avgs)[1..50] );
    $stats{'median'} = $stat->median();
    $stats{'stddev'} = $stat->standard_deviation();
    $stats{'mean'} = $stat->mean();
    $stats{'min'} = $stat->min();
    $stats{'max'} = $stat->max();

    return \%stats;
}


sub parse_line
{
    my $line = shift;

    my @fields = grep { defined $_ } ($line =~ m{ \[([^\]]+)\] | "([^"]+)" | (\S+) }gx);

    my $access = {
        time => $fields[3],
        request => $fields[4],
        agent => $fields[8]
    };

    $access->{agent} = $1 if $access->{agent} =~ m{((?:Firefox|Opera|Batik)/[\d\.]+|Batik/ batik-[\d\._]+)};
    $access->{time} = $1 if $access->{time} =~ m{:(\d\d:\d\d:\d\d)};
    $access->{request} =~ s/^(GET|POST) //;
    $access->{request} =~ s{ HTTP/[\d\.]+$}{};

    return $access;
}

=head1 NAME

generate_chart.pl - Generate charts from SVG profile log files.

=head1 VERSION

This document describes generate_chart.pl version 0.0.1

=head1 SYNOPSIS

  generate_chart.pl [options] {logfiles}

=head1 DESCRIPTION

The C<generate_chart.pl> program extracts information from an Apache log file
to chart the results of a performance test using the F<ProfilerLogger.es>
profiling tool.

=head2 COMMAND LINE OPTIONS

The following command line options control the function on the program.

=over 4

=item --app={application name}

This selects which application to extract from the log files. The application
is determined from the C<marker> parameter passed to the F<logit.html> file.
This parameter is required for the program to function.

=item --outfile={SVG filename}

This option selects the name of the file where the SVG chart is written. If
this parameter is not supplied the chart is written to STDOUT.

=back

The program also expects a list of Apache log files that are parsed to
retrieve the information used to build the charts.
=head1 DIAGNOSTICS

=for author to fill in:
    List every single error and warning message that the module can
    generate (even the ones that will "never happen"), with a full
    explanation of each problem, one or more likely causes, and any
    suggested remedies.

=over

=item C<< Error message here, perhaps with %s placeholders >>

[Description of error here]

=item C<< Another error message here >>

[Description of error here]

[Et cetera, et cetera]

=back


=head1 DEPENDENCIES

The program depends on C<Carp>, C<Statistics::Descriptive>, C<Getopt::Long>,
C<Pod::Usage> and C<Template>

=head1 INCOMPATIBILITIES

None reported.

=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to the author

=head1 AUTHOR

G. Wade Johnson  C<< wade@anomaly.org >>

=head1 LICENCE AND COPYRIGHT

Copyright (c) 2008, G. Wade Johnson C<< wade@anomaly.org >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.

=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.

=cut

__DATA__
<svg height="[%height%]" width="[%width%]" viewBox="0 0 [%width%] [%height%]"
     xmlns="http://www.w3.org/2000/svg"
     xmlns:xlink="http://www.w3.org/1999/xlink">
<!--
   File:        [% filename +%]
   Author:      G. Wade Johnson

   Copyright: 2009 G. Wade Johnson
      You are free to use this code however you like.
 -->
  <title>Chart of [% app %] Average Performance</title>
  <desc>This chart shows the mean performance of the [% app %] SVG application
    as a function of update interval size in milliseconds.
  </desc>

  <defs>
    <style type="text/css">
        .heading {
            text-anchor: middle;
            font-family: arial, helvetica, sans-serif;
            font-size: 16px;
        }
        .labels { fill: black; font-size:10px; }
        .title {
            text-anchor: middle;
            font-family: arial, helvetica, sans-serif;
            font-size: 12px;
        }
        .detail, #instructions {
            font-family: arial, helvetica, sans-serif;
            font-size: 10px;
        }
        .rtjust { text-anchor: end; }
        .center { text-anchor: middle; }
        .axes {
            stroke: black;
            stroke-width: 1px;
        }
        .grid {
            stroke: #888;
            stroke-width: 0.5px;
        }
        .background {
            fill: #eef;
            stroke: none;
        }
        .legend { font-size: 12px; fill: black; }
        .legend_bg {
            stroke: black;
            stroke-width: 0.5px;
            fill: #eee;
        }
[% FOREACH i = agent_indexes %]
        .data[% i %], .marker[% i %] {
            fill: none;
            stroke: [% color(i) %];
            stroke-width: 1px;
        }
        .marker[% i %] {
            fill: [% color(i) %];
            fill-opacity: 0.6;
            fill-rule: evenodd;
            stroke-width: 0.5px;
        }
        .legend[% i %] {
            fill: [% color(i) %];
            stroke: none;
        }
[% END %]
    </style>
    <script type="text/ecmascript"><![CDATA[
        function show( el )
        {
            el.setAttributeNS( null, 'visibility', 'visible' )
        }
        function hide( el )
        {
            el.setAttributeNS( null, 'visibility', 'hidden' )
        }
        function is_hidden( el )
        {
            return 'hidden' == el.getAttributeNS( null, 'visibility' );
        }
        function showdata( n )
        {
            for(var i=0; i < [%agent_count%]; ++i)
            {
                var g = document.getElementById( 'points'+i );
                if(null == g)
                    return;
                if(n == i && is_hidden( g ))
                    show( g );
                else
                    hide( g );
            }
        }
    ]]></script>
  </defs>
  <rect x="0" y="0" width="[%width%]" height="[%height%]" class="background"/>

  <text x="[% heading.x %]" y="[% heading.y %]" class="heading">[% app %] (mean)</text>
  <path class="grid" d="[% FOREACH x = xpoints %] M[%x.x%],[%orig.y%]v4 L[%x.x%],[%top.y%][% END %]"/>
  <line x1="[% orig.x %]" y1="[% orig.y %]" x2="[% orig.x %]" y2="[% top.y %]" class="axes"/>
  <line x1="[% orig.x %]" y1="[% orig.y %]" x2="[% top.x %]" y2="[% orig.y %]" class="axes"/>
  <text x="[% xtitle.x %]" y="[% xtitle.y %]" class="title">Interval Size (ms)</text>
  <text x="[% ytitle.x %]" y="[% ytitle.y %]" class="title" transform="rotate(-90,[%ytitle.x%],[%ytitle.y%])">Updates/Sec</text>
[% FOREACH x = xpoints %]
  <text x="[% x.x %]" y="[% xlabel_y %]" class="center labels">[% x.p %]</text>
[% END %]
[% FOREACH y = ypoints %]
  <text x="[% ylabel_x %]" y="[% y.y %]" class="rtjust labels">[% y.l %]</text>
[% END %]
  <path class="grid" d="[% FOREACH y = ypoints %] M[%orig.x%],[%y.y%]h-4 L[%top.x%],[%y.y%][% END %]"/>

  <g transform="translate(0,[%orig.y%])">
[% FOREACH i = agent_indexes %]
    <path class="data[%i%]" d="[% data_line( agent(i) ) -%]"/>
[% FOREACH e = error_bars( agent(i) ) -%]
    <path class="marker[%i%]" d="[%e%]"/>
[% END %]
[% END %]
[% FOREACH i = agent_indexes %]
    <g id="points[%i%]" class="detail" visibility="hidden">
[% FOREACH p = data_detail( agent(i) ) -%]
      <g transform="translate( [%p.x%], [%p.y%] )">
        <rect width="[%p.width%]" height="[%p.height%]" stroke="[%color(i)%]" stroke-width="1" fill="#fff"/>
        <text x="3" y="12">Interval: [%p.interval%]</text>
        <text x="3" y="24">Mean: [%p.mean%]</text>
        <text x="3" y="36">Min: [%p.min%]</text>
        <text x="3" y="48">Max: [%p.max%]</text>
        <text x="3" y="60">StdDev: [%p.stddev%]</text>
      </g>
[% END %]
    </g>
[% END %]
  </g>
  <g transform="translate([%legend_size.x%],[%legend_size.y%])"
     onload="show( document.getElementById('instructions') )">
    <rect x="0" y="0" width="[%legend_size.width%]" height="[%legend_size.height%]" class="legend_bg"/>
[% FOREACH i = agent_indexes %]
    <rect x="5" y="[%legend(i).block%]" width="10" height="10" class="legend[%i%]" onclick="showdata([%i%])"/>
    <text x="20" y="[%legend(i).text%]" class="legend">[%agent(i)%]</text>
[% END %]
    <text x="-15" y="[%legend_size.height+12%]" id="instructions" visibility="hidden">Click colored blocks for detail</text>
  </g>
</svg>
